
#ifndef _BASE_LIST_HXX_
#define _BASE_LIST_HXX_ 1

/*
 *	Copyright (C)  by
 *	ines GmbH 1991
 */



#ifndef _BASE_LIST_HXX
#define	_BASE_LIST_HXX 1

typedef unsigned uint;

class ChainRaw;

/*
 *	LIST NOMENCLATURE

	Link:	+---+	+---+   +---+
		  n  ->   n  ->   n\
		 \p  <-   p  <-   p
		+---+	+---+   +---+
	Chain:	^		^
		head		tail

	[ direction: head to tail ]

	put(el, nel)		append nel to el 
	put(0, nel)		insert nel @head
	put(tail(), nel) 	append to list [push]
	
	remove(el)		remove elem from list
	remove(head())		remove head element [pull]
	remove(tail())		remove tail element [pop]
 */

/*
 *	a Link 
 *
 *	if a Link can be on different chains, and this is significant.
 *	deletion of a Link will remove it automatic from it's chain.
 */

class Link {
	friend class ChainRaw;
	friend class Chain;

	ChainRaw *_chain;
	Link *_next;
	Link *_prev;
public:
	Link();				// construct isolated (!) Link
	Link(ChainRaw *);		// construct a Link, append it to chain
	virtual ~Link();		// remove also from chain
	ChainRaw* chain() const { return _chain; }
	int isIsol() { return (_chain==0); }
					// get chain or null
	Link *next() { return _next; }
	Link *prev() { return _prev; }
};

struct ChainIterator {
	Link*	_pliCur;	/* current during iteration */
	int	_fTH;		/* iterate for tail to head if set */
};
#define ChainIter ChainIterator 

class ChainRaw {
	Link *_head;
	Link *_tail;
	uint  _c;

public:
	ChainRaw(); 			// construct
	ChainRaw(ChainRaw*);		// imports source chain, doesn't duplicate links!
	~ChainRaw();			// doesn't delete links

	/*
	 *	start iteration
	 *	you can iterate from head to tail or reversed
	 */
	void start_iter(ChainIterator *pit)	/* obsolete */
		const { pit->_pliCur = _head; pit->_fTH=0; }
	void strtIter(ChainIterator *pit) const   { start_iter(pit); }
	void strtIterHT(ChainIterator *pit) const { start_iter(pit); }
	void strtIterTH(ChainIterator *pit) const { pit->_pliCur=_tail; pit->_fTH=1; }

	/*
	 *	iterate
	 *	returns 0 if iteration done. 
	 *	Direction is given by startIter??.
	 *	save during iteration:
	 *		Deletion/removal of the current node
	 *	Usage:
	 *
	 *	strtIter(&it)
	 *	while (pl=iterate(&it)) {
	 *		If (...) delete(pl);
	 *	}
	 */
	Link* iterate(ChainIterator*) const;	
	Link* iter(ChainIterator* pit)	const { return iterate(pit); }

	/*
	 *	peek()
	 *	returns the next link to be returned by iter() without
	 *	moving forward/backward. 
	 *	Nodes appended( during head/tail iter ) to peek() 
	 *	are included in the iteration. 
	 *	Deletion/removal of peek() is not allowed. 
	 *
	 *	Usage (a): Append during iteration
	 *
	 *	strtIter(&it)
	 *	while (pl=peek(&it)) {
	 *		If (...) append(pl, new Link());
	 *		(void)iter();
	 *	}
	 *
	 *	Usage (b): Test for list end.
	 *
	 *	strtIter(&it)
	 *	while (pl=iter(&it)) {
	 *		cout << (*pl);
	 *		if (peek()) cout << ','; // not at end
	 *		(void)iter();
	 *	}
 	 *	
	 */
	Link *peek(ChainIterator *) const ;

	Link* head() const { return _head; } 		// first elem or null
	Link* tail() const { return _tail; }
	Link* put(Link*l, Link*nel);	// put nel behind 'l', ret nel or 0
	Link* append(Link*l, Link*nel) { return put(l, nel); }
	Link* insert(Link*l, Link*nel) { return put(l ? l->prev() : 0, nel); }

	Link* remove(Link*);		// remove Link from list,fatal if not
	Link* push(Link *l)		// append to chain, fatal if on oth ch
		{ return put(_tail, l); }
	Link  *pop()			// remove last elem
		{ return remove(_tail); }
	Link  *pull()			// remove first elem
		{ return remove(_head); }
	uint count() const { return _c; }	// count links on chain
	uint c() const { return count(); }	// new name 

	/*  read chain to array */
	int get(Link **, int c);
	Link *get(int iPos) const;
	int pos(const Link *pln) const;

	/*
	 *	import()
	 *	remove from other chain, append to this
	 */
	Link* import(Link *pln) {	
		if (pln->chain()) pln->chain()->remove(pln);
		this->push(pln);
		return pln;
	}
	void import(ChainRaw *pcn);	/* return this	*/
};


/*
 *	CHAINITER macro
 *	simplifies coding of chain iterations
 *	usage:
 *	
 *	class X:public Link { int _y; };
 *	Chain ch;
 *	CHAINITER( ch, X*, pX) {
 *		 if (pX->_y) ...
 *	}
 */
#define CHAINITER( chname, linkcast, linkvar ) \
	ChainIterator cit;\
	linkcast linkvar;\
	chname.strtIter(&cit);\
	while(linkvar=(linkcast)chname.iter(&cit))
#endif

/*
 *	Lightweight Lists
 *
 *	- no explicit chain object
 *	- no recursive deletion
 */
class LiLink {
	LiLink *_pllNext;
	LiLink *_pllPrev;
public:
	LiLink();
	~LiLink();
	
	LiLink *pllNext() { return _pllNext; }
	LiLink *pllPrev() { return _pllPrev; }

	int isHead() { return _pllPrev == 0; }
	int isTail() { return _pllNext == 0; }
	int isIsol() { return ((!_pllNext) && (!_pllPrev)); }


	LiLink *append(LiLink *pllNew);		/* after this	*/
	LiLink *insert(LiLink *pllNew);		/* before this	*/
	LiLink *remove();
};

/*
 *	Lightweight Chain - no recursive deletion
 */
class LiChain {
	LiLink *_pllHead;
	LiLink *_pllTail;
public:
	LiChain();
	~LiChain();			/* no deletion */
	
	LiLink *appendTail(LiLink *pll);/* append to tail */
	LiLink *removeHead();		/* link or 0	*/
	LiLink *remove(LiLink *pll);		/* link or 0	*/
};

/*
 *	Lightweight List with payload
 */
class LiLiPl:public LiLink {
	void *_pv;
public:
	LiLiPl(void *p);		/* use LiLink if mounting not required */
	~LiLiPl();

	void *pv() { return _pv ; }
	void *pvNext() { return pllNext() ? ((LiLiPl*)pllNext())->_pv : 0; }
	void *pvPrev() { return pllPrev() ? ((LiLiPl*)pllPrev())->_pv : 0; }
};

#endif
