#ifndef _GRULS_HXX
#define _GRULS_HXX 1

  /**
   * GRand Unified Logging System
   */

namespace INdepGruls100 {

  enum { ceMaxAttr=4, ceMaxPath=16 };
  union Attr {
    int _i;
    char *_sz;
  };

  struct Attrs {
    Attr _mpat [ceMaxAttr];	/* indexed by key */
  };

  class Logger;
  class Core;

  /** Logging Record */
  class Record {
   friend class Core;
   friend class Logger;
   const Logger *_plo;
   int _iLvl;
   char _mpch[256];		/* the buffer */
   public:
    int iLvl() const { return _iLvl; }
    const Logger* plo() const { return _plo; }
    const char* pch() const { return _mpch; }
    Record *puts(const char *);	/**< write the logging message */
    Record *printf(const char *, ...);	
  };

  typedef int atin_t;

  /*
   * logging severity levels
   *
   * 0 off
   * 1 least severe
   * 9 most severe
   */


  class Core {
    int _iLogDefault;		/* default return of getConfig() */
   public:
    Record _rc;

    virtual Record *openRecordLvl(const Logger *plo, int iLvl);	/* check logging for combined attribute set */
    virtual void flushRecord(Record *prc);		/* done logging, send to sink		*/
    virtual void init(Logger *);			/**< register logger   */
    virtual void fini(Logger *);			/**< unregister logger */
    virtual int getConfig(const char **ppch, const char **ppchLim);
    inline Record *openRecord(const Logger *plo, int iLvl);
    Core(int iLogDefault=0xf):_iLogDefault(iLogDefault) {}
    ~Core() { }
  };


  /** Logger */
  class Logger {
    friend class Core;
    const char *_szId;	/* instance id */
    Core *_pco;		/* logging core reference */
    Attrs _ats;		/* this logger specific attributes */
    const Logger *_ploPar;	/* parent logger */

    /* properties of Core, initialized by Core during registration */
    int _fMask;		/* logging flag, set by Core::init()	*/
    void *_phd;		/* core handle after registration (optional)  */

    const char** getIdPath(const char **ppchBuf, const char** ppchLim) const ;
   public:
    const Logger *ploPar() const { return _ploPar; }
    const char *szId() const { return _szId; }
    /** set Core properties, access for Core only */
    void setMask(int f, void *p=0) { _fMask=f; _phd=p; }

    void setAttr(atin_t, int i);	
    void setAttr(atin_t, const char *sz);	/* sz not saved or released */

    /*
     * inline record open 
     */
    Record *openRecord(int iLvl) const { return _pco ? _pco->openRecord(this, iLvl) : 0 ; } 
    void flushRecord(Record *prc) const { return _pco->flushRecord(prc); } 

    /** 
     * scope loggers, extend parent by attribute(s) 
     * 
     * \note The (static) Core must be constructed and _configured_ before any
     * Logger is constructed. 
     */
    Logger(const char *szId, const Logger *ploPar=0);
    ~Logger();
    //explicit Logger(const Logger &);
    //Logger(const Logger *ploPar);
    //Logger(const Logger *ploPar, atin_t, int i);	/* child logger with single integer attribute */
    //Logger(const Logger *ploPar, atin_t, char *sz);	/* child logger with single string attribute */

  };

  inline Record *Core::openRecord(const Logger *plo, int fLog) {
    /* predefined (quick) log level check */
    if (fLog & plo->_fMask) return openRecordLvl(plo, fLog);
    else return 0;
  }

  extern Core *pGrulsCore;	/* to be initialized before any use */
}

  /* generic logging flags */
  #define LOG_FAIL 0x8 // 1000
  #define LOG_LESS 0x4 // 0100
  #define LOG_MORE 0x2	// 0010
  #define LOG_VERB 0x1 // 0001

  /* generic escalated logging mask */
  #define LOM_FAIL 0x8	// x000
  #define LOM_LESS 0xb // xx00
  #define LOM_MORE 0xe // xxx0
  #define LOM_VERB 0xf // xxxx


  #define LOGS(lg, lvl,  string) \
  if (INdepGruls100::Record *prc=(lg).openRecord((lvl))) (lg).flushRecord(prc->puts(string)); 

  #define LOGF(lg, lvl,  format, ...) \
  if (INdepGruls100::Record *prc=(lg).openRecord((lvl))) (lg).flushRecord(prc->printf(format, __VA_ARGS__)); 

#ifndef _MM_NO_IOSTREAM

  /*
   * extension for systems using iostreams
   */
  #include <sstream>

  #define LOGOS(lg, lvl,  exp) \
  if (INdepGruls100::Record *prc=(lg).openRecord((lvl))) { \
      std::ostringstream oss; \
      oss << exp << std::flush; \
      prc->puts(oss.str().c_str()); \
      (lg).flushRecord(prc); \
  } 

  #define LOGOF(lg, lvl,  func) \
  if (INdepGruls100::Record *prc=(lg).openRecord((lvl))) { \
      std::ostringstream oss; \
      func; \
      oss << std::flush; \
      prc->puts(oss.str().c_str()); \
      (lg).flushRecord(prc); \
  } 
#endif

#endif
