/**
 * \file 
 * Timer/Counter (TIC) C Interface
 */

/**
 * \name Status functions
 */
/**@{*/

/** 
 * get the most error status. 
 * Error details may requested by the tic_errortext(), tic_errorcode and tic_diagcode() functions.
 *
 * \param i device index, 0 for the first device attached
 * \return nonzero if failed, zero otherwise.
 */
extern   int tic_failed(int i) ; 	

/** 
 * Error message of the most recent error encountered. 
 *
 * \param i device index, 0 for the first device attached
 * \return error message string
 */
extern   const char *tic_errortext(int i) ;

/** 
 * Error code of the most recent error encountered. 
 *
 * \param i device index, 0 for the first device attached
 * \return error code (hexadecimal)
 */
extern   int tic_errorcode(int i) ;

/** 
 * Diagnostic code of the most recent error encountered. 
 *
 * An additional diagnostic code may be given for some errors.
 *
 * \param i device index, 0 for the first device attached
 * \return diagnostic code 
 */
extern   int tic_diagcode(int i) ; 	
/**@}*/

/**
 * \name Lifecycle functions
 */
/**@{*/

/** 
 * initialize device 
 *
 * \param i device index, 0 for the first device attached
 * \return nonzero if failed, zero otherwise.
 */
extern   int tic_init(int i) ;

/** 
 * finish device use
 *
 * \param i device index, 0 for the first device attached
 * \return nonzero if failed, zero otherwise.
 */
extern   int tic_fini(int i) ;
/**@}*/


/**
 * \name Measurement functions
 */
/**@{*/

/** 
 * sample count
 *
 * \param i device index, 0 for the first device attached
 * \return  sample count, negative if  failed.
 */
extern   int tic_count_samples(int i) ;

/** 
 * main loop, controls the device operation during measurements
 *
 * \param i device index, 0 for the first device attached
 * \return nonzero if failed, zero otherwise.
 */
extern   int tic_loop(int i) ;

/**@}*/

/**
 * \name MFAV functions
 */
/**@{*/

/** 
 * set MFAV (frequeny average mode)
 *
 * \param i device index, 0 for the first device attached
 * \return nonzero if failed, zero otherwise.
 */
extern   int tic_mode_frequency_average(int i) ;

/** 
 * read sample
 *
 * \param i device index, 0 for the first device attached
 * \return  sample value, (-1.0 if  failed).
 */
extern   double tic_read_frequency_average(int i) ;
/**@}*/

/**
 * \name MPWI functions
 */
/**@{*/

/** 
 * set MPWI (pulse width mode)
 *
 * \param i device index, 0 for the first device attached
 * \return nonzero if failed, zero otherwise.
 */
extern   int tic_mode_pulse_width(int i) ;

/** 
 * read sample
 *
 * \param i device index, 0 for the first device attached
 * \return  sample value, (-1.0 if  failed).
 */
extern   double tic_read_pulse_width(int i) ;
/**@}*/

/**
 * \name MPUL functions
 */
/**@{*/

/** 
 * set MPUL (pulse mode)
 *
 * \param i device index, 0 for the first device attached
 * \return nonzero if failed, zero otherwise.
 */
extern   int tic_mode_pulse(int i) ;

/** 
 * read first half of sample
 *
 * \param i device index, 0 for the first device attached
 * \return  width of the first half of the pulse, (-1.0 if  failed).
 */
extern   double tic_read_pulse_first(int i) ;

/** 
 * read second half of sample
 *
 * \param i device index, 0 for the first device attached
 * \return  width of the second half of the pulse, (-1.0 if  failed).
 */
extern   double tic_read_pulse_second(int i) ;
/**@}*/

/**
 * \name MTOC functions
 */
/**@{*/

/** 
 * set MTOC (totalizing counter)
 *
 * \param i device index, 0 for the first device attached
 * \return nonzero if failed, zero otherwise.
 */
extern   int tic_mode_totalcount(int i) ;

/** 
 * read totalized count
 *
 * \param i device index, 0 for the first device attached
 * \return  totalized count
 */
extern   long long tic_read_totalcount(int i) ;

/**@}*/


/**
 * \name Control functions
 */
/**@{*/
/** 
 * arm channel
 *
 * \param i device index, 0 for the first device attached
 * \return nonzero if failed, zero otherwise.
 */
extern int tic_arm(int i) ;

/** 
 * set mode properties
 *
 * \param i device index, 0 for the first device attached
 * \param eProp property key
 * \param iVal property value
 * \return nonzero if failed, zero otherwise.
 */
extern int tic_set_property(int i, unsigned eProp, int iVal);
/**@}*/

/**
 * \name Test/Demo functions
 */
/**@{*/
/**
 * enable 200 kHz test mode stimuli at port IO15
 * \param i device index, 0 for the first device attached
 * \warning Do not attach external signals when using test mode.
 */
extern   int tic_init_stimuli(int i) ;
/**@}*/
