
/**
 * \file
 * Umi Task and Functions
 *
 * an implementation of cooperative multi-tasking on top of the Universal Message Interface.
 */

#ifndef _UMI_UMI_TASK_HXX_
#define _UMI_UMI_TASK_HXX_

#include "umic/umi_driver.hxx"
#include "umi/umi_user.hxx"
#include <stack>
#include <assert.h>

namespace INdepUmi100 {


  class TaskFlags {
    int _f;

   public:
    int f() const { return _f; }
    void f(int __f) { _f=__f; }

    int fCanget(int i) const { return (1 << i) << 0; }  /* down data in (get, p</assert>md()) available (0xf) */
    int fCanput(int i) const { return (1 << i) << 4; }  /* up data (put, pmu()) available (0xf0)     */
    int fSendup(int i) const { return (1 << i) << 8; }   /* send upstream (0xf00)  */
    int fReleasedown(int i) const { return (1 << i) << 12; } /* 0xf000 */
    int fCanget2(int i) const { return (1 << i) << 16; }  /* down data in (get, pmd(0+1)) available 0xf0000 */
    int fUser() const { return 0x100000; }  /* down data in (get, pmd(0+1)) available 0xf0000 */

    int isSync(int f) { return (f & 0xf00ff) == (_f & 0xf00ff); }

    int isCanget(int i) const { return _f & ((1 << i) << 0); }  
    int isCanput(int i) const { return _f & ((1 << i) << 4); }  
    int isSendup(int i) const { return _f & ((1 << i) << 8); }  
    int isReleasedown(int i) const { return _f & ((1 << i) << 12); }  
    int isCanget2(int i) const { return _f & ((1 << i) << 16); }  

    int isCanget() const { return _f & ((0xf) << 0); }  
    int isCanput() const { return _f & ((0xf) << 4); }  
    int isSendup() const { return _f & ((0xf) << 8); }  
    int isReleasedown() const { return _f & ((0xf) << 12); }  
    int isCanget2() const { return _f & ((0xf) << 16); }  

    TaskFlags(int f=0):_f(f) {}
    ~TaskFlags() { }
  };



  /* 
   * function base class 
   */
  class Task;

  class TaskUser { 
   public:
    virtual void done(Task *pta);
    TaskUser() {}
    ~TaskUser() {}
  };


  class Task {
    TaskUser *_ptm;
    int _h;
    Channel *_mppcn[2];
   public:
    TaskUser *ptm() const { return _ptm; }
    int h() const { return _h; }		/* handle at ptm	*/
    Channel *pcn(int i) const { return _mppcn[i]; }
    virtual void sigDown(int iCtx, Channel *pcn) { }
    virtual void sigUp(int iCtx, Channel *pcn) { }
    Task(TaskUser *ptm, int h, Channel *pcn0, Channel *pcn1):_ptm(ptm), _h(h) {  
      _mppcn[0]=pcn0;
      _mppcn[1]=pcn1;
    }
    virtual ~Task() { }
  };

  class TaskForward:public Task {
    int _c;
   public:
    void sigDown(int iCtx, Channel *pcn);
    void sigUp(int iCtx, Channel *pcn);
    TaskForward(TaskUser *ptm, int h, Channel *pcn0, Channel *pcn1);
    ~TaskForward();
  };

  /**
   * Task that runs Functions
   *
   * calls parent done() on return
   */
  class TaskFunc:public Task {
   public:

    /**
     * base class for user defined functions 
     */
    class Func:public TaskFlags {
      friend class TaskFunc;
      TaskFunc *_ptt;
      int _e;	  /* "program counter" state */
    protected:
      static Time timestamp();	/* REMIND: not wrap save, must reboot within 136 years */

      TaskFunc *ptt() const { return _ptt; }
      int e() const { return _e; }

      Pmu pmu(int iChan, int iOffs=0) const;
      Pmd pmd(int iChan, int iOffs=0) const;
      Channel *pcn(int i) const;

      int call(int sNext, Func *ptf);	/* call a sub function */

      /* 
       * setup in a single call:
       *
       * - postprocessing to be done after return from func()
       * - preconditions for the next call to func()
       */
      int setup(int sNext, int fPre, int fPost=0) { 
	_e=sNext;
	f(fPre|fPost);
	return f();
      }

     public:

      /** 
       * user defined non-blocking function
       */
      virtual int func( int eState )=0;

     public:
      void init(TaskFunc *ptt) { _ptt=ptt; }
      Func():_ptt(0), _e(0) {  
      }
      virtual ~Func() { }
    };

    class Frame {
      Func *_ptf;
     public:
      Func *ptf() { return _ptf; }
      Frame& operator=(const Frame &fr) { _ptf=fr._ptf; return *this; }
      Frame(Func *ptf): _ptf(ptf) { }
      Frame(): _ptf(0) { }
      ~Frame() { }
    };

   private:
    std::stack<Frame> _stfr;
    int _cRec;
    Func *_ptfRet;		/* most recent returned function */


    int cChan() const { return 2; }
    void setPtfRet(Func *ptf) { if (_ptfRet && ptf != _ptfRet) delete _ptfRet; _ptfRet=ptf; }
    //void setPtfRet(Func *ptf) { delete _ptfRet; _ptfRet=ptf; }


   protected:
   public:
    void run(Func *ptf);
    Func *ptfRet() { return _ptfRet; } /* most recent returned func */

    Pmu pmu(int iChan, int iOffs=0) const { return pcn(iChan)->pmuUp(iOffs); } 
    Pmd pmd(int iChan, int iOffs=0) const { return pcn(iChan)->pmdDown(iOffs); }



    void sigDown(int iCtx, Channel *pcn); 
    void sigUp(int iCtx, Channel *pcn);
    void proceed(int iCtx , std::stack<Frame> &stfr );
    int call(Func *ptf);

    bool isRunning() { return !_stfr.empty(); }


    TaskFunc(TaskUser *ptm, int h, Channel *pcn0, Channel *pcn1, Func *ptf=0);
    ~TaskFunc();
  };

  inline Pmu TaskFunc::Func::pmu(int iChan, int iOffs) const { return ptt()->pmu(iChan, iOffs); } 
  inline Pmu TaskFunc::Func::pmd(int iChan, int iOffs) const { return ptt()->pmd(iChan, iOffs); } 
  inline Channel *TaskFunc::Func::pcn(int i) const { return ptt()->pcn(i); }
  inline int TaskFunc::Func::call(int sNext, TaskFunc::Func *ptf) { _e=sNext; return ptt()->call(ptf); }

  /**
   * single channel (client) plexer with task
   *
   * drives a single process when sigDown/sigUp is called (driven e.g. by poll())
   */
  class TaskPlex:public Plexer, public TaskFunc, public TaskUser {

  protected:
    

   public:

    /* implement Plexer */
    void sigDown(int iCtx, Channel *pcn, int hChan) { TaskFunc::sigDown(iCtx, pcn); }
    void sigUp(int iCtx, Channel *pcn, int hChan) { TaskFunc::sigUp(iCtx, pcn); }

    /** connects to puc (initPlex) during lifetime. */
    TaskPlex( UmiChannel *puc ):Plexer("tp"), TaskFunc(this, 0, puc, 0) {
      pcn(0)->initPlex(this,0);
    }
    ~TaskPlex( ) {
      pcn(0)->finiPlex();
    }
  };

  /**
   * Plexer multi-channel with Ports
   *
   * Tasks are created dynamically
   */
  class PortMux:public TaskUser,  public Plexer {
   protected:
    struct PortClass {
      const char *_szPortId;
      const Proto *_pup;
    };

    class Port {
      Channel *_pcn;	/* client channel */
      Task *_pta;
     public:
      Task *pta() const { return _pta; }
      Channel *pcn() const { return _pcn; }
      Channel *setChan(Channel *pcn) { return _pcn=pcn; }
      void setTask(Task *pta) { _pta=pta; }
      void clrTask() { _pta=0; }
      Port(Channel *pcn):_pcn(pcn), _pta(0) {}
      Port():_pcn(0), _pta(0) {}
      ~Port() {}
    };

   private:
    const PortClass *_pPc;
    Port *_pPo;		/* all ports */
    Port _poDev;	/* device pseudo-port */
    int cPortTypes();


   public:

    /* implement configuration */
    int openChannel( const char *szPort, Channel *pud );
    void closeChannel( int iPort );

   protected:

    Channel *pucDev() { return _poDev.pcn(); }
    /* implement Umi::Plexer */
    void sigDown(int iCtx, Channel *pcn, int hChan);
    void sigUp(int iCtx, Channel *pcn, int hChan);

    /**
     * create task at port
     */
    virtual Task *createTask(Channel *, const PortClass *, Pmd pmd, PortMux *pma, int h)=0;
    virtual void deleteTask(const PortClass*, Task *pta)=0;

    void done(Task *pts);	/* task done */

    
    /**
     * The PortClass (const) and Port linear arrays are to be supplied by the derived class.
     * 
     * pcnDev - if set, managed by the PortMux
     */
    PortMux(const PortClass *pPc, Port *pPo, Channel *pcnDev=0);
    ~PortMux();
  };

}


#endif
