
#include "umic/umi.h"
#include "mod.h"
#include "sizes.h"
#include "card.h"
#include "assert.h"
#include <linux/fs.h>
#include <asm/uaccess.h>

#include <linux/version.h>
#if ( LINUX_VERSION_CODE < (KERNEL_VERSION( 2,5,0 )) )
#include <linux/sched.h>
#else
#include <linux/interrupt.h>
#endif

  void pca_lock(Card *pca) { 
    spin_lock(&pca->_sl);
  }

  int pca_unlock(Card *pca, int e) { 
    spin_unlock(&pca->_sl);
    return e;
  }

  #define pca_failed(pca) ((pca)->_fState == SCA_FAIL) 
  #define pca_idle(pca) ((pca)->_fState == SCA_IDLE) 
  #define pca_iDriver(pca) ((pca)->_iDriver) 
  #define pca_pcnByChanOfCard(pca, iChanOfCard) ((pca)->_mpcn+(iChanOfCard))


  int pca_initDriver(Card *pca, int iDriver) {
    pca->_iDriver = iDriver;
    pca->_fState = SCA_INIT;
    pca->_pfInit(pca, iDriver);
    return 0;
  }

  int pca_finiDriver(Card *pca) {
    pca->_pfFini(pca, pca->_iDriver);
    pca->_iDriver = -1;
    pca->_fState = SCA_IDLE;
    return 0;
  }

  int pca_checkidle(Card *pca) {
    Chan *pcn;
    for (pcn=pca->_mpcn; pcn < pca->_mpcn+MAX_CACN; pcn++) {
      if (pcn_isopen(pcn)) return 0;
    }
    return 1;
  }

  const struct UmiProto *pcn_pup(Chan *pcn) { return pcn->_pUdd->_pup; }

  #define pcn_pdm(pcn) (pcn->_pca->_pdm)
  #define pcn_pca(pcn) (pcn->_pca)
  #define pcn_debug(pcn, str, ...) \
   if (isDebug(pcn_pdm(pcn), pcn->_pca)) \
     printk(KERN_DEBUG PFX "%s: " str "\n", (pcn_pdm(pcn))->_szDid, pdm_getid(pcn_pdm(pcn), pcn->_pca), ##__VA_ARGS__ )

  #define pca_debug(pca, str, ...) \
   if (isDebug(pca_pdm(pca), pca)) \
     printk(KERN_DEBUG PFX "%s: " str "\n", (pca_pdm(pca))->_szDid, pdm_getid(pca_pdm(pca), pca), __VA_ARGS__ )



  unsigned int umi_poll(struct file *fp, poll_table *pt ) {
    int i;
    Chan *pcn;
    pcn=(Chan*)(fp->private_data);
    assert(pcn);


    i=pcn_poll(pcn, fp, pt);
    //printk(KERN_INFO PFX "poll %x\n", pcn_pdm(pcn)->_szDid, i);
    return i;
  }

  long umi_unlocked_ioctl(struct file *fp, unsigned int io_cmd, unsigned long arg ) {
    unsigned int cmd=_IOC_NR(io_cmd);
    UmiIoc *pio=(void*)arg;
    Chan *pcn=(Chan*)(fp->private_data);
    assert(pcn);

    pcn_debug(pcn,  "ioctl cmd=%x pcn=%p", _IOC_NR(cmd), pcn);

    if (!access_ok(VERIFY_WRITE, arg, sizeof(UmiIoc))) return -EFAULT;

    if (cmd == GIOC_PROCHK) {
      /* check the protocol before exchanging message */
      unsigned char* pchBuf;
      int e=0;
      size_t s;
      pchBuf =  kmalloc (SZ_PROTID, GFP_KERNEL);
      assert(pchBuf);

      /* get and check the protocol descriptor */
      if (bufPutDown(&pio->_dd, pchBuf, SZ_PROTID, &s)) { e= -EFAULT; goto lDone; }
      if (!UmiProto_match(pcn_pup(pcn), pchBuf, strlen(pchBuf))) { e = -EIO; goto lDone; }

     lDone:
      kfree(pchBuf);
      return e;
    }


    /*
     * we must be owner of the channel 
     * to use the following commands
     */
    if (1) { 

      /*
       * this thread now owns pcn and pcn->_puc
       * access to the channels putdown and getup functions
       * will be single threaded.
       */
      if (cmd & GIOC_PUTMSG) {
	int iErr;

	/*
	 * serialize access among concurrent threads
	 * using the same channel/mode. EAGAIN is returned
	 * to the thread trying to access a channel/mode
	 * which is already executed.
	 */
	if (pcn_lockPut(pcn)) {
	  iErr=pcn_putdown(pcn, pio);
	  if (iErr) {
	    pcn_debug( pcn, "putdown %i", iErr);
	    return pcn_unlockPut(pcn, iErr);
	  }
	  pcn_debug( pcn, "putdown ok");
	  pcn_stat(pcn, "ioctl-put-done");
	  (void)pcn_unlockPut(pcn, 0);
	}
	else return -EAGAIN;
      }

      if (cmd & GIOC_GETMSG) {
	  int iErr;

	if (pcn_lockGet(pcn)) {

	  pcn_stat(pcn, "ioctl-get-ok");
	  iErr=pcn_getup(pcn, pio);
	  if (iErr) {
	    pcn_debug( pcn, "getup %d", iErr);
	    return pcn_unlockGet(pcn, iErr);
	  }
	  pcn_debug( pcn, "getup ok");

	  pcn_stat(pcn, "ioctl-get-done");
	  (void)pcn_unlockGet(pcn,0);
	}
	else return -EAGAIN;
	pcn_stat(pcn, 0);
      }
      pcn_debug( pcn, "ioctl-done");
    }
    else return -EBUSY;

    return 0;
  }

  /**
   * open a UMI channel
   */
  int umi_open(LumiMod *pdm, struct inode * pin, struct file *fp) {
    int iChanOfCard;
    int iDriverOfCard;
    int iChanOfDriver;
    int iCard;
    Card *pca;
    Chan *pcn;

    iCard=GETCARD(MINOR(pin->i_rdev));
    iChanOfCard=GETCHAN(MINOR(pin->i_rdev));
    iDriverOfCard=GETDRIV(MINOR(pin->i_rdev));
    iChanOfDriver=GETDRCN(MINOR(pin->i_rdev));

    /*
     * check if the card is available. This is concurrent
     * to the hotplug probing process.
     */
    printk(KERN_DEBUG PFX " open ca=%d, cc=%d, dr=%d, cd=%d\n", pdm->_szDid, iCard, iChanOfCard, iDriverOfCard, iChanOfDriver);
    if (iCard < 0 || iCard >= MAX_CARD) return -ENODEV;
    if (iChanOfCard < 0 || iChanOfCard >= MAX_CACN) return -ENODEV;

    pca=pdm_getcard(pdm, iCard);
    if (!pca) return -ENODEV;



    /* card state switch */

    pca_lock(pca);

    if (pca_idle(pca)) {
      /*
       * first open of any driver channel
       * define the card's driver and mode
       */
      pca_debug(pca, "init driver %d", iDriverOfCard);
      if (pca_initDriver(pca, iDriverOfCard)) return pca_unlock(pca, -EIO);
    }
    else if (pca_failed(pca)) {
      return pca_unlock(pca, -EIO);
    }
    assert(!pca_idle(pca));
    assert(!pca_failed(pca));

    /*
     * check if the card is already 
     * in use by another driver
     */
    if (pca_iDriver(pca) != iDriverOfCard) return pca_unlock(pca, -EBUSY);


    /* driver valid and initialized */
    pcn=pca_pcnByChanOfCard(pca, iChanOfCard);
    if (!pcn_open(pcn)) return pca_unlock(pca,-EBUSY);

    pca_debug(pca, "chan=%p", pcn);


    fp->private_data=pcn;
    pcn_debug( pcn, "open %d/%d", MAJOR(pin->i_rdev), MINOR(pin->i_rdev));
    return pca_unlock(pca,0);
  }

  int umi_release( struct inode * pin, struct file *fp) {
    Card *pca;
    Chan *pcn=fp->private_data;
    pcn_debug( pcn, "close %d", pin->i_rdev);
    pca=pcn_pca(pcn);
    pcn_close(pcn);

    pca_lock(pca);
    if (pca_checkidle(pca)) {
      pca_finiDriver(pca);
      pcn_debug(pcn, "fini %d", 0  );
    }
    return pca_unlock(pca, 0);
  }
